"""
Provider Index

This module contains a list of all available data providers.
"""

from datetime import date, timedelta

from meteostat.typing import License, ProviderSpec
from meteostat.enumerations import Grade, Granularity, Parameter, Provider, Priority

PROVIDER_HOURLY = ProviderSpec(
    id=Provider.HOURLY,
    name="Meteostat Hourly",
    granularity=Granularity.HOURLY,
    priority=Priority.NONE,
    grade=None,
    license=None,
    parameters=[
        Parameter.TEMP,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.SNWD,
        Parameter.WDIR,
        Parameter.WSPD,
        Parameter.WPGT,
        Parameter.PRES,
        Parameter.TSUN,
        Parameter.CLDC,
        Parameter.COCO,
    ],
    start=date(1931, 1, 1),
    module="meteostat.providers.meteostat.hourly",
)

PROVIDER_DAILY = ProviderSpec(
    id=Provider.DAILY,
    name="Meteostat Daily",
    granularity=Granularity.DAILY,
    priority=Priority.NONE,
    grade=None,
    license=None,
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.SNWD,
        Parameter.WSPD,
        Parameter.WPGT,
        Parameter.PRES,
        Parameter.TSUN,
        Parameter.CLDC,
    ],
    start=date(1899, 1, 1),
    module="meteostat.providers.meteostat.daily",
)

PROVIDER_DAILY_DERIVED = ProviderSpec(
    id=Provider.DAILY_DERIVED,
    name="Meteostat Daily Derived",
    granularity=Granularity.DAILY,
    priority=Priority.NONE,
    grade=None,
    license=None,
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.SNOW,
        Parameter.WSPD,
        Parameter.WDIR,
        Parameter.WPGT,
        Parameter.PRES,
        Parameter.TSUN,
        Parameter.CLDC,
    ],
    start=date(1931, 1, 1),
    module="meteostat.providers.meteostat.daily_derived",
)

PROVIDER_MONTHLY = ProviderSpec(
    id=Provider.MONTHLY,
    name="Meteostat Monthly",
    granularity=Granularity.MONTHLY,
    priority=Priority.NONE,
    grade=None,
    license=None,
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.PRES,
        Parameter.TSUN,
    ],
    start=date(1899, 1, 1),
    module="meteostat.providers.meteostat.monthly",
)

PROVIDER_MONTHLY_DERIVED = ProviderSpec(
    id=Provider.MONTHLY_DERIVED,
    name="Meteostat Monthly Derived",
    granularity=Granularity.MONTHLY,
    priority=Priority.NONE,
    grade=None,
    license=None,
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.PRES,
        Parameter.TSUN,
    ],
    start=date(1899, 1, 1),
    module="meteostat.providers.meteostat.monthly_derived",
)

PROVIDER_DWD_HOURLY = ProviderSpec(
    id=Provider.DWD_HOURLY,
    name="DWD Hourly",
    granularity=Granularity.HOURLY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="Deutscher Wetterdienst",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    countries=["DE"],
    parameters=[
        Parameter.TEMP,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.WDIR,
        Parameter.WSPD,
        Parameter.PRES,
        Parameter.TSUN,
        Parameter.VSBY,
        Parameter.COCO,
    ],
    start=date(1939, 1, 1),
    module="meteostat.providers.dwd.hourly",
)

PROVIDER_DWD_POI = ProviderSpec(
    id=Provider.DWD_POI,
    name="DWD POI",
    granularity=Granularity.HOURLY,
    priority=Priority.MEDIUM,
    grade=Grade.OBSERVATION,
    license=License(
        commercial=True,
        attribution="Deutscher Wetterdienst",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    parameters=[
        Parameter.CLDC,
        Parameter.TEMP,
        Parameter.VSBY,
        Parameter.WSPD,
        Parameter.WPGT,
        Parameter.PRCP,
        Parameter.RHUM,
        Parameter.WDIR,
        Parameter.PRES,
        Parameter.TSUN,
        Parameter.COCO,
        Parameter.SNWD,
    ],
    start=date.today() - timedelta(days=1),
    module="meteostat.providers.dwd.poi",
)

PROVIDER_DWD_MOSMIX = ProviderSpec(
    id=Provider.DWD_MOSMIX,
    name="DWD MOSMIX",
    granularity=Granularity.HOURLY,
    priority=Priority.LOWEST,
    grade=Grade.FORECAST,
    license=License(
        commercial=True,
        attribution="Deutscher Wetterdienst",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    parameters=[
        Parameter.TEMP,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.WDIR,
        Parameter.WSPD,
        Parameter.WPGT,
        Parameter.PRES,
        Parameter.TSUN,
        Parameter.CLDC,
        Parameter.COCO,
    ],
    start=date.today(),
    module="meteostat.providers.dwd.mosmix",
)

PROVIDER_DWD_DAILY = ProviderSpec(
    id=Provider.DWD_DAILY,
    name="DWD Daily",
    granularity=Granularity.DAILY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="Deutscher Wetterdienst",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    countries=["DE"],
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.SNWD,
        Parameter.WSPD,
        Parameter.WPGT,
        Parameter.PRES,
        Parameter.TSUN,
        Parameter.CLDC,
    ],
    start=date(1781, 1, 1),
    module="meteostat.providers.dwd.daily",
)

PROVIDER_DWD_MONTHLY = ProviderSpec(
    id=Provider.DWD_MONTHLY,
    name="DWD Monthly",
    granularity=Granularity.MONTHLY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="Deutscher Wetterdienst",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    countries=["DE"],
    parameters=[Parameter.TEMP, Parameter.TMIN, Parameter.TMAX, Parameter.PRCP],
    start=date(1851, 1, 1),
    module="meteostat.providers.dwd.monthly",
)

PROVIDER_ECCC_HOURLY = ProviderSpec(
    id=Provider.ECCC_HOURLY,
    name="ECCC Hourly",
    granularity=Granularity.HOURLY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="Environment and Climate Change Canada",
        name="Environment and Climate Change Canada Data Servers End-use Licence",
        url="https://eccc-msc.github.io/open-data/licence/readme_en/",
    ),
    countries=["CA"],
    parameters=[
        Parameter.TEMP,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.WDIR,
        Parameter.WSPD,
        Parameter.VSBY,
    ],
    start=date(1939, 1, 1),
    module="meteostat.providers.eccc.hourly",
)

PROVIDER_ECCC_DAILY = ProviderSpec(
    id=Provider.ECCC_DAILY,
    name="ECCC Daily",
    granularity=Granularity.DAILY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="Environment and Climate Change Canada",
        name="Environment and Climate Change Canada Data Servers End-use Licence",
        url="https://eccc-msc.github.io/open-data/licence/readme_en/",
    ),
    countries=["CA"],
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.SNOW,
        Parameter.SNWD,
        Parameter.WPGT,
    ],
    start=date(1939, 1, 1),
    module="meteostat.providers.eccc.daily",
)

PROVIDER_ECCC_MONTHLY = ProviderSpec(
    id=Provider.ECCC_MONTHLY,
    name="ECCC Monthly",
    granularity=Granularity.MONTHLY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="Environment and Climate Change Canada",
        name="Environment and Climate Change Canada Data Servers End-use Licence",
        url="https://eccc-msc.github.io/open-data/licence/readme_en/",
    ),
    countries=["CA"],
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.PDAY,
        Parameter.SNOW,
    ],
    start=date(1939, 1, 1),
    module="meteostat.providers.eccc.monthly",
)

PROVIDER_ISD_LITE = ProviderSpec(
    id=Provider.ISD_LITE,
    name="NOAA ISD Lite",
    granularity=Granularity.HOURLY,
    priority=Priority.HIGH,
    grade=Grade.OBSERVATION,
    license=License(
        commercial=True,
        name="CC0 1.0 Universal",
        url="https://creativecommons.org/publicdomain/zero/1.0/",
    ),  # source: https://registry.opendata.aws/noaa-isd/
    parameters=[
        Parameter.TEMP,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.WDIR,
        Parameter.WSPD,
        Parameter.WPGT,
        Parameter.PRES,
        Parameter.CLDC,
    ],
    start=date(1931, 1, 1),
    module="meteostat.providers.noaa.isd_lite",
)

PROVIDER_GHCND = ProviderSpec(
    id=Provider.GHCND,
    name="NOAA GHCN-Daily",
    granularity=Granularity.DAILY,
    priority=Priority.HIGH,
    grade=Grade.OBSERVATION,
    license=License(
        commercial=True,
        name="CC0 1.0 Universal",
        url="https://creativecommons.org/publicdomain/zero/1.0/",
    ),  # source: https://registry.opendata.aws/noaa-ghcn/
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.SNOW,
        Parameter.WSPD,
        Parameter.WDIR,
        Parameter.WPGT,
        Parameter.TSUN,
        Parameter.CLDC,
    ],
    start=date(1931, 1, 1),
    module="meteostat.providers.noaa.ghcnd",
)

PROVIDER_CLIMAT = ProviderSpec(
    id=Provider.CLIMAT,
    name="DWD CLIMAT",
    granularity=Granularity.MONTHLY,
    priority=Priority.HIGH,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="Deutscher Wetterdienst",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.TXMN,
        Parameter.TXMX,
        Parameter.PRES,
        Parameter.PRCP,
        Parameter.TSUN,
    ],
    start=date(1875, 6, 1),
    module="meteostat.providers.dwd.climat",
)

PROVIDER_METAR = ProviderSpec(
    id=Provider.METAR,
    name="METAR",
    granularity=Granularity.HOURLY,
    priority=Priority.LOW,
    grade=Grade.OBSERVATION,
    license=License(
        commercial=True,
    ),
    parameters=[
        Parameter.TEMP,
        Parameter.RHUM,
        Parameter.WDIR,
        Parameter.WSPD,
        Parameter.PRES,
        Parameter.COCO,
    ],
    start=date.today() - timedelta(days=1),
    module="meteostat.providers.noaa.metar",
)

PROVIDER_METNO_FORECAST = ProviderSpec(
    id=Provider.METNO_FORECAST,
    name="MET Norway Forecast",
    granularity=Granularity.HOURLY,
    priority=Priority.LOWEST - 1,  # MOSMIX is more specific, therefore higher prio
    grade=Grade.FORECAST,
    license=License(
        commercial=True,
        attribution="MET Norway",
        name="NLOD 2.0",
        url="https://data.norge.no/nlod/en/2.0",
    ),
    parameters=[
        Parameter.TEMP,
        Parameter.RHUM,
        Parameter.PRCP,
        Parameter.WDIR,
        Parameter.WSPD,
        Parameter.WPGT,
        Parameter.PRES,
        Parameter.CLDC,
        Parameter.COCO,
    ],
    start=date.today(),
    module="meteostat.providers.metno.forecast",
)

PROVIDER_GSA_HOURLY = ProviderSpec(
    id=Provider.GSA_HOURLY,
    name="GeoSphere Austria Hourly",
    granularity=Granularity.HOURLY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="GeoSphere Austria",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    countries=["AT"],
    parameters=[
        Parameter.TEMP,
        Parameter.PRCP,
        Parameter.PRES,
        Parameter.WSPD,
        Parameter.WDIR,
        Parameter.RHUM,
        Parameter.TSUN,
    ],
    start=date(1880, 4, 1),
    module="meteostat.providers.gsa.hourly",
)

PROVIDER_GSA_SYNOP = ProviderSpec(
    id=Provider.GSA_SYNOP,
    name="GeoSphere Austria SYNOP",
    granularity=Granularity.HOURLY,
    priority=Priority.HIGH,
    grade=Grade.OBSERVATION,
    license=License(
        commercial=True,
        attribution="GeoSphere Austria",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    countries=["AT"],
    parameters=[
        Parameter.TEMP,
        Parameter.PRES,
        Parameter.RHUM,
        Parameter.WSPD,
        Parameter.WDIR,
        Parameter.PRCP,
    ],
    start=date(2000, 1, 1),
    module="meteostat.providers.gsa.synop",
)

PROVIDER_GSA_DAILY = ProviderSpec(
    id=Provider.GSA_DAILY,
    name="GeoSphere Austria Daily",
    granularity=Granularity.DAILY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="GeoSphere Austria",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    countries=["AT"],
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.PRES,
        Parameter.RHUM,
        Parameter.WSPD,
        Parameter.TSUN,
    ],
    start=date(1880, 4, 1),
    module="meteostat.providers.gsa.daily",
)

PROVIDER_GSA_MONTHLY = ProviderSpec(
    id=Provider.GSA_MONTHLY,
    name="GeoSphere Austria Monthly",
    granularity=Granularity.MONTHLY,
    priority=Priority.HIGHEST,
    grade=Grade.RECORD,
    license=License(
        commercial=True,
        attribution="GeoSphere Austria",
        name="CC BY 4.0",
        url="https://creativecommons.org/licenses/by/4.0/",
    ),
    countries=["AT"],
    parameters=[
        Parameter.TEMP,
        Parameter.TMIN,
        Parameter.TMAX,
        Parameter.PRCP,
        Parameter.PRES,
        Parameter.TSUN,
    ],
    start=date(1880, 4, 1),
    module="meteostat.providers.gsa.monthly",
)


DEFAULT_PROVIDERS = [
    PROVIDER_HOURLY,
    PROVIDER_DAILY,
    PROVIDER_DAILY_DERIVED,
    PROVIDER_MONTHLY,
    PROVIDER_MONTHLY_DERIVED,
    PROVIDER_DWD_HOURLY,
    PROVIDER_DWD_POI,
    PROVIDER_DWD_MOSMIX,
    PROVIDER_DWD_DAILY,
    PROVIDER_DWD_MONTHLY,
    PROVIDER_ECCC_HOURLY,
    PROVIDER_ECCC_DAILY,
    PROVIDER_ECCC_MONTHLY,
    PROVIDER_ISD_LITE,
    PROVIDER_GHCND,
    PROVIDER_CLIMAT,
    PROVIDER_METAR,
    PROVIDER_METNO_FORECAST,
    PROVIDER_GSA_HOURLY,
    PROVIDER_GSA_SYNOP,
    PROVIDER_GSA_DAILY,
    PROVIDER_GSA_MONTHLY,
]



"""
Inventory Module

Provides classes for working with weather station data inventories.
"""

from datetime import date, datetime
from typing import List, Optional

import pandas as pd

from meteostat.enumerations import Parameter


class Inventory:
    """
    A weather station's data inventory
    """

    df: Optional[pd.DataFrame] = None

    def __init__(self, df: Optional[pd.DataFrame] = None):
        if df is not None and not df.empty:
            self.df = df

    @property
    def start(self) -> Optional[date]:
        """
        Get the earliest start date from the inventory
        """
        return (
            datetime.strptime(self.df["start"].min(), "%Y-%m-%d").date()
            if self.df is not None
            else None
        )

    @property
    def end(self) -> Optional[date]:
        """
        Get the latest end date from the inventory
        """
        return (
            datetime.strptime(self.df["end"].max(), "%Y-%m-%d").date()
            if self.df is not None
            else None
        )

    @property
    def parameters(self) -> Optional[List[Parameter]]:
        """
        Get the list of available parameters from the inventory
        """
        if self.df is None:
            return []

        return [
            Parameter[parameter.upper()]
            for parameter in self.df.index.get_level_values("parameter").unique()
        ]



"""
The code is licensed under the MIT license.
"""

from datetime import datetime
from typing import Optional

import pandas as pd

from meteostat.providers.meteostat.shared import filter_model_data, handle_exceptions
from meteostat.typing import ProviderRequest
from meteostat.api.config import config
from meteostat.core.cache import cache_service
from meteostat.utils.data import reshape_by_source


ENDPOINT = config.hourly_endpoint


def get_ttl(_station: str, year: int) -> int:
    """
    Get TTL based on year

    Current + previous year = one day
    Else = 30 days
    """
    current_year = datetime.now().year
    return 60 * 60 * 24 if current_year - year < 2 else 60 * 60 * 24 * 30


@cache_service.cache(get_ttl, "pickle")
@handle_exceptions
def get_df(station: str, year: int) -> Optional[pd.DataFrame]:
    """
    Get CSV file from Meteostat and convert to DataFrame
    """
    file_url = ENDPOINT.format(station=station, year=str(year))

    df = pd.read_csv(file_url, sep=",", compression="gzip")

    time_cols = df.columns[0:4]
    df["time"] = pd.to_datetime(df[time_cols])
    df = df.drop(time_cols, axis=1).set_index("time")

    return reshape_by_source(df)


@filter_model_data
def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    """
    Fetch hourly weather data from Meteostat's central data repository
    """
    if req.start is None or req.end is None:
        return None

    # Get a list of relevant years
    years = range(req.start.year, req.end.year + 1)
    # Get list of annual DataFrames
    df_yearly = [get_df(req.station.id, year) for year in years]
    # Concatenate into a single DataFrame
    df = (
        pd.concat(df_yearly)
        if len(df_yearly) and not all(d is None for d in df_yearly)
        else None
    )
    return df



from typing import Optional, Union
from urllib.error import HTTPError

import pandas as pd

from meteostat.core.cache import cache_service
from meteostat.core.logger import logger
from meteostat.enumerations import TTL, Parameter
from meteostat.typing import ProviderRequest
from meteostat.utils.conversions import percentage_to_okta

ENDPOINT = "https://opendata.dwd.de/weather/weather_reports/poi/{station}-BEOB.csv"
# TODO: add col 11 for solar radiation support (Globalstrahlung)
USECOLS = [0, 1, 2, 9, 14, 21, 22, 23, 33, 35, 36, 37, 40, 41]
NAMES = {
    "Wolkenbedeckung": Parameter.CLDC,
    "Temperatur (2m)": Parameter.TEMP,
    "Sichtweite": Parameter.VSBY,
    "Windgeschwindigkeit": Parameter.WSPD,
    "Windboen (letzte Stunde)": Parameter.WPGT,
    "Niederschlag (letzte Stunde)": Parameter.PRCP,
    "Relative Feuchte": Parameter.RHUM,
    "Windrichtung": Parameter.WDIR,
    "Druck (auf Meereshoehe)": Parameter.PRES,
    "Sonnenscheindauer (letzte Stunde)": Parameter.TSUN,
    "aktuelles Wetter": Parameter.COCO,
    "Schneehoehe": Parameter.SNWD,
}
COCO_MAP = {
    "1": 1,
    "2": 2,
    "3": 3,
    "4": 4,
    "5": 5,
    "6": 6,
    "7": 7,
    "8": 8,
    "9": 9,
    "10": 10,
    "11": 11,
    "12": 12,
    "13": 13,
    "14": 14,
    "15": 15,
    "16": 16,
    "17": 24,
    "18": 17,
    "19": 18,
    "20": 19,
    "21": 20,
    "22": 21,
    "23": 22,
    "24": 19,
    "25": 20,
    "26": 23,
    "27": 25,
    "28": 26,
    "29": 25,
    "30": 26,
    "31": 27,
}


def get_coco(code: str | int) -> Union[int, None]:
    """
    Map DWD POI weather condition codes to Meteostat condicodes
    """
    return COCO_MAP.get(str(code))


@cache_service.cache(TTL.HOUR, "pickle")
def get_df(station: str) -> Optional[pd.DataFrame]:
    try:
        # Read CSV data from DWD server
        df = pd.read_csv(  # type: ignore
            ENDPOINT.format(station=station),
            sep=";",
            skiprows=2,
            na_values="---",
            usecols=USECOLS,
            decimal=",",
        )

        # Rename columns
        df = df.rename(columns=NAMES)

        # Snow cm -> mm
        df[Parameter.SNWD] = df[Parameter.SNWD].multiply(10)
        df[Parameter.VSBY] = df[Parameter.VSBY].multiply(1000)

        # Change coco
        df[Parameter.COCO] = df[Parameter.COCO].apply(get_coco)
        df[Parameter.CLDC] = df[Parameter.CLDC].apply(percentage_to_okta)

        # Set index
        df["time"] = pd.to_datetime(
            df["Datum"] + " " + df["Uhrzeit (UTC)"], format="%d.%m.%y %H:%M"
        )
        df = df.set_index(["time"])
        df = df.drop(["Datum", "Uhrzeit (UTC)"], axis=1)

        return df

    except HTTPError as error:
        logger.info(
            f"Couldn't load DWD POI data for weather station {station} (status: {error.status})"
        )
        return None

    except Exception as error:
        logger.warning(error)
        return None


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    if "wmo" in req.station.identifiers:
        return get_df(req.station.identifiers["wmo"])



from typing import Optional

from meteostat.enumerations import TTL
from meteostat.core.cache import cache_service
from meteostat.core.network import network_service
from meteostat.core.logger import logger


ENDPOINT = "https://api.weather.gc.ca"


@cache_service.cache(TTL.WEEK)
def get_meta_data(station: str) -> Optional[dict]:
    response = network_service.get(
        f"{ENDPOINT}/collections/climate-stations/items",
        params={
            "STN_ID": station,
            "properties": ",".join(
                [
                    "CLIMATE_IDENTIFIER",
                    "TIMEZONE",
                    "HLY_FIRST_DATE",
                    "HLY_LAST_DATE",
                    "DLY_FIRST_DATE",
                    "DLY_LAST_DATE",
                    "MLY_FIRST_DATE",
                    "MLY_LAST_DATE",
                ]
            ),
            "f": "json",
        },
    )

    if response.status_code == 200:
        data = response.json()
        try:
            return data["features"][0]["properties"]
        except (IndexError, KeyError):
            logger.info(f"ECCC climate identifier for station {station} not found")
    else:
        logger.warning(
            f"ECCC climate identifier for station {station} not found (status: {response.status_code})"
        )

    return None



"""
█▀▄▀█ █▀▀ ▀█▀ █▀▀ █▀█ █▀ ▀█▀ ▄▀█ ▀█▀
█░▀░█ ██▄ ░█░ ██▄ █▄█ ▄█ ░█░ █▀█ ░█░

A Python library for accessing open weather and climate data.

Meteorological data provided by Meteostat (https://dev.meteostat.net)
under the terms of the Creative Commons Attribution 4.0 International
License.

The code is licensed under the MIT license.
"""

__appname__ = "meteostat"
__version__ = "2.1.2"

from meteostat.api.daily import daily
from meteostat.api.hourly import hourly
from meteostat.api.interpolate import interpolate
from meteostat.api.inventory import Inventory
from meteostat.api.merge import merge
from meteostat.api.monthly import monthly
from meteostat.api.normals import normals
from meteostat.api.point import Point
from meteostat.api.stations import stations
from meteostat.api.timeseries import TimeSeries
from meteostat.core.cache import purge
from meteostat.api.config import config
from meteostat.enumerations import Granularity, Parameter, Provider, UnitSystem
from meteostat.interpolation.lapserate import lapse_rate
from meteostat.typing import Station, License

# Export public API
__all__ = [
    "config",
    "daily",
    "hourly",
    "Granularity",
    "interpolate",
    "Inventory",
    "lapse_rate",
    "License",
    "merge",
    "monthly",
    "normals",
    "Parameter",
    "Point",
    "Provider",
    "purge",
    "Station",
    "stations",
    "TimeSeries",
    "UnitSystem",
]



"""
Network Service

The Network Service provides methods to send HTTP requests
considering the Meteostat configuration.
"""

from typing import Optional

import requests

from meteostat import __version__
from meteostat.core.logger import logger
from meteostat.api.config import config


class NetworkService:
    """
    Network Service
    """

    @staticmethod
    def _process_headers(headers: dict) -> dict:
        """
        Process headers
        """

        headers["X-Meteostat-Version"] = __version__

        return headers

    def get(
        self,
        url: str,
        params=None,
        headers: Optional[dict] = None,
        stream: Optional[bool] = None,
    ) -> requests.Response:
        """
        Send a GET request using the Meteostat configuration
        """
        if headers is None:
            headers = {}

        headers = self._process_headers(headers)

        return requests.get(
            url,
            params,
            headers=headers,
            stream=stream,
            proxies=config.network_proxies,
            timeout=30,
        )

    def get_from_mirrors(
        self,
        mirrors: list[str],
        params=None,
        headers: Optional[dict] = None,
        stream: Optional[bool] = None,
    ) -> Optional[requests.Response]:
        """
        Send a GET request to multiple mirrors using the Meteostat configuration
        """
        for mirror in mirrors:
            try:
                response = self.get(
                    mirror,
                    params=params,
                    headers=headers,
                    stream=stream,
                )
                if response.status_code == 200:
                    return response
            except requests.RequestException:
                logger.warning("Could not fetch data from '%s'", mirror)
                continue
        return None


network_service = NetworkService()



"""
The code is licensed under the MIT license.
"""

from typing import Any, Optional

import pandas as pd
from metar import Metar

from meteostat.core.logger import logger
from meteostat.api.config import config
from meteostat.enumerations import TTL, Frequency, Parameter
from meteostat.typing import ProviderRequest
from meteostat.utils.conversions import temp_dwpt_to_rhum
from meteostat.core.cache import cache_service
from meteostat.utils.data import enforce_freq
from meteostat.core.network import network_service


ENDPOINT = config.aviationweather_endpoint
USER_AGENT = config.aviationweather_user_agent
CLDC_MAP = {
    "FEW": 2,  # 1-2 octas
    "SCT": 4,  # 3-4 octas
    "BKN": 6,  # 5-7 octas
    "OVC": 8,  # 8 octas (fully overcast)
}
COCO_MAP = {
    "RA": 8,
    "SHRA": 17,
    "DZ": 7,
    "DZRA": 7,
    "FZRA": 10,
    "FZDZ": 10,
    "RASN": 12,
    "SN": 15,
    "SHSN": 21,
    "SG": 12,
    "IC": 12,
    "PL": 24,
    "GR": 24,
    "GS": 24,
    "FG": 5,
    "BR": 5,
    "MIFG": 5,
    "BCFG": 5,
    "HZ": 5,
    "TS": 25,
    "TSRA": 25,
    "PO": 27,
    "SQ": 27,
    "FC": 27,
    "SS": 27,
    "DS": 27,
}


def safe_get(obj: Optional[Any]) -> Optional[Any]:
    try:
        if obj is not None and hasattr(obj, "value"):
            return obj.value()
        return None
    except AttributeError:
        return None


def get_cldc(report: Metar.Metar) -> Optional[int]:
    """
    Get cloud cover (octas) from METAR report
    """
    try:
        cloud_cover = report.sky[0][0]
        return CLDC_MAP.get(cloud_cover)
    except IndexError:
        return None


def get_coco(report: Metar.Metar) -> Optional[int]:
    """
    Get weather condition code from METAR report
    """
    try:
        condition_code = "".join(
            [item for item in report.weather[0] if item is not None]
        )
        return COCO_MAP.get(condition_code)
    except IndexError:
        return None


def map_data(record):
    """
    Map METAR data to Meteostat column names
    """
    try:
        parsed_report = Metar.Metar(record)

        return {
            "time": parsed_report.time,
            Parameter.TEMP: safe_get(parsed_report.temp),
            Parameter.DWPT: safe_get(parsed_report.dewpt),
            Parameter.PRCP: safe_get(parsed_report.precip_1hr),
            Parameter.SNWD: safe_get(parsed_report.snowdepth),
            Parameter.WDIR: safe_get(parsed_report.wind_dir),
            Parameter.WSPD: safe_get(parsed_report.wind_speed),
            Parameter.WPGT: safe_get(parsed_report.wind_gust),
            Parameter.PRES: safe_get(parsed_report.press),
            Parameter.VSBY: safe_get(parsed_report.vis),
            Parameter.CLDC: get_cldc(parsed_report),
            Parameter.COCO: get_coco(parsed_report),
        }
    except Metar.ParserError:
        return None


@cache_service.cache(TTL.HOUR, "pickle")
def get_df(station: str) -> Optional[pd.DataFrame]:
    """
    Get CSV file from Meteostat and convert to DataFrame
    """
    url = ENDPOINT.format(station=station)

    if not USER_AGENT:
        logger.warning(
            "Consider specifying a unique user agent when querying the Aviation Weather Center's data API."
        )

    headers = {"User-Agent": USER_AGENT}

    response = network_service.get(url, headers=headers)

    # Raise an exception if the request was unsuccessful
    response.raise_for_status()

    # Parse the JSON content into a DataFrame
    data = [
        item for item in map(map_data, response.text.splitlines()) if item is not None
    ]

    # Return None if no data is available
    if not len(data):
        return None

    # Create DataFrame
    df = pd.DataFrame(data)

    # Return None if DataFrame is empty
    if df.empty:
        return None

    # Add RHUM column
    df[Parameter.RHUM] = df.apply(lambda row: temp_dwpt_to_rhum(row), axis=1)
    df[Parameter.RHUM] = pd.to_numeric(df[Parameter.RHUM], errors="coerce").round()

    # Set time index
    df = df.set_index(["time"])

    return enforce_freq(df, Frequency.HOURLY)


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    if "icao" in req.station.identifiers:
        return get_df(req.station.identifiers["icao"])



"""
Stations Module

Provides the Stations class for working with weather station metadata.
"""

import math
import os
from typing import List, Optional
from io import BytesIO
import sqlite3

import pandas as pd

from requests import Response
from meteostat.api.inventory import Inventory
from meteostat.api.point import Point
from meteostat.core.cache import cache_service
from meteostat.api.config import config
from meteostat.core.logger import logger
from meteostat.core.network import network_service
from meteostat.enumerations import Provider
from meteostat.typing import Station


class Stations:
    """
    Stations Database
    """

    def _register_math_functions(self, conn: sqlite3.Connection) -> None:
        """
        Register mathematical functions for SQLite.

        SQLite doesn't always have built-in math functions (depends on compile options).
        This method registers Python implementations to ensure compatibility.
        """
        # Register trigonometric functions used in distance calculations
        conn.create_function("acos", 1, math.acos)
        conn.create_function("cos", 1, math.cos)
        conn.create_function("sin", 1, math.sin)

        # Register radians/degrees conversion
        conn.create_function("radians", 1, math.radians)
        conn.create_function("degrees", 1, math.degrees)

    def _fetch_file(self, stream=False) -> Response:
        """
        Download the SQLite database file from the configured URL
        """
        urls = config.stations_db_endpoints

        if not urls:
            raise Exception("No stations database URLs configured")

        response = network_service.get_from_mirrors(urls, stream=stream)

        if response is None:
            raise Exception("Failed to download the database file")

        return response

    def _get_file_path(self) -> str:
        """
        Get the file path for the SQLite database
        """
        filepath = config.stations_db_file
        ttl = config.stations_db_ttl

        if os.path.exists(filepath) and not cache_service.is_stale(filepath, ttl):
            return filepath

        # Download the database file
        response = self._fetch_file(stream=True)

        # Create cache directory if it doesn't exist
        cache_service.create_cache_dir()

        with open(filepath, "wb") as file:
            for chunk in response.iter_content(chunk_size=8192):
                file.write(chunk)

        return filepath

    def _connect_memory(self) -> sqlite3.Connection:
        """
        Create an in-memory SQLite database and load the downloaded database file into it
        """
        # Download the database file
        response = self._fetch_file()

        # Create an in-memory SQLite database
        conn = sqlite3.connect(":memory:")

        # Register math functions for compatibility
        self._register_math_functions(conn)

        # Read the downloaded database file into memory
        content = BytesIO(response.content)

        # Convert bytes to string and write the content to the in-memory database
        conn.deserialize(content.read())

        return conn

    def _connect_fs(self) -> sqlite3.Connection:
        """
        Connect to the SQLite database file on the filesystem
        """
        file = self._get_file_path()

        if not file:
            raise FileNotFoundError("SQLite database file not found")

        conn = sqlite3.connect(file)

        # Register math functions for compatibility
        self._register_math_functions(conn)

        return conn

    def connect(self, in_memory: Optional[bool] = None) -> sqlite3.Connection:
        """
        Connect to the database
        """
        if in_memory is None:
            in_memory = config.stations_db_file is None

        logger.info("Connecting to stations database (in_memory=%s)", in_memory)

        if in_memory:
            return self._connect_memory()

        return self._connect_fs()

    def query(
        self,
        sql: str,
        index_col: Optional[str | list] = None,
        params: Optional[tuple | dict] = None,
    ) -> pd.DataFrame:
        """
        Execute a SQL query and return the result as a DataFrame
        """
        with self.connect() as conn:
            df = pd.read_sql(sql, conn, index_col=index_col, params=params)

        return df

    def meta(self, station: str) -> Station:
        """
        Get meta data for a specific weather station
        """
        meta = self.query(
            """
            SELECT 
                `stations`.*,
                `names`.`name` as `name`
            FROM `stations` 
            LEFT JOIN `names` ON `stations`.`id` = `names`.`station` 
                AND `names`.`language` = 'en'
            WHERE `stations`.`id` LIKE ?
            """,
            index_col="id",
            params=(station,),
        ).to_dict("records")[0]

        identifiers = self.query(
            "SELECT `key`, `value` FROM `identifiers` WHERE `station` LIKE ?",
            params=(station,),
        ).to_dict("records")

        return Station(
            id=station,
            **meta,
            identifiers={
                identifier["key"]: identifier["value"] for identifier in identifiers
            },
        )

    def inventory(
        self, station: str | List[str], providers: Optional[List[Provider]] = None
    ) -> Inventory:
        """
        Get inventory records for a single weather station
        """
        query = "SELECT station, provider, parameter, start, end, completeness FROM `inventory`"
        station_list = station if isinstance(station, list) else [station]

        # Generate the right number of placeholders (?, ?, ?, ...)
        placeholders = ", ".join(["?"] * len(station_list))
        # Add the placeholders to the query
        query += f" WHERE `station`  IN ({placeholders})"
        # Add the stations to the params
        params = tuple(station_list)

        if providers:
            # Generate the right number of placeholders (?, ?, ?, ...)
            placeholders = ", ".join(["?"] * len(providers))
            # Add the placeholders to the query
            query += f" AND provider IN ({placeholders})"
            # Add the providers to the params
            params += tuple(providers)

        df = self.query(
            query, index_col=["station", "provider", "parameter"], params=params
        )

        return Inventory(df)

    def nearby(self, point: Point, radius=50000, limit=100) -> pd.DataFrame:
        """
        Get a list of weather station IDs ordered by distance
        """
        return self.query(
            """
            SELECT
                `id`,
                `names`.`name` as `name`,
                `country`,
                `region`,
                `latitude`,
                `longitude`,
                `elevation`,
                `timezone`,
                ROUND(
                    (
                        6371000 * acos(
                            cos(radians(:lat)) * cos(radians(`latitude`)) * 
                            cos(radians(`longitude`) - radians(:lon)) + 
                            sin(radians(:lat)) * sin(radians(`latitude`))
                        )
                    ),
                    1
                ) AS `distance`
            FROM
                `stations`
                INNER JOIN `names` ON `stations`.`id` = `names`.`station`
                AND `names`.`language` = "en"
            WHERE
                `distance` <= :radius
            ORDER BY
                `distance`
            LIMIT
                :limit
            """,
            index_col="id",
            params={
                "lat": point.latitude,
                "lon": point.longitude,
                "radius": radius,
                "limit": limit,
            },
        )


stations = Stations()



"""
Schema Service

The Schema Service provides methods to clean and format
DataFrames based on a set of parameters.
"""

from copy import copy
from inspect import isfunction
from typing import Callable, List

import pandas as pd

from meteostat.core.logger import logger
from meteostat.core.parameters import parameter_service
from meteostat.core.validator import Validator
from meteostat.enumerations import Granularity, Parameter, UnitSystem
from meteostat.utils.conversions import CONVERSION_MAPPINGS, to_condition, to_direction


class SchemaService:
    """
    Schema service
    """

    @staticmethod
    def _apply_validator(
        validator: Validator | Callable, df: pd.DataFrame, col: str
    ) -> pd.Series:
        """
        Apply a validator
        """
        if isfunction(validator):
            v: Validator = validator()  # type: ignore
        else:
            v = validator

        if not isinstance(v, Validator):
            return pd.Series(data=True, index=df.index, dtype=bool)

        if v.ignore_na:
            result = pd.Series(data=True, index=df.index, dtype=bool)
            test_result = v.test(
                df.loc[df[col].notnull()][col],
                df.loc[df[col].notnull()],
                col,
            )
            if isinstance(test_result, pd.Series):
                result.update(test_result)
            return result.astype(bool)

        test_result = v.test(df[col], df, col)
        if isinstance(test_result, bool):
            return pd.Series(data=test_result, index=df.index, dtype=bool)
        return test_result

    @staticmethod
    def purge(df: pd.DataFrame, parameters: List[Parameter]) -> pd.DataFrame:
        """
        Remove DataFrame columns which are not a known parameter
        """
        columns = [parameter for parameter in parameters if parameter in df.columns]
        return df[columns]

    @staticmethod
    def fill(df: pd.DataFrame, parameters: List[Parameter]) -> pd.DataFrame:
        """
        Add missing schema columns to DataFrame
        """
        for parameter_id in parameters:
            if parameter_id not in df:
                df[parameter_id] = None

        return df

    @staticmethod
    def format(df: pd.DataFrame, granularity: Granularity) -> pd.DataFrame:
        """
        Set data types and round values
        """
        temp = copy(df)

        for col in df.columns:
            parameter = parameter_service.get_parameter(col, granularity)

            if not parameter:
                logger.warning(
                    "Column %s is not a valid column name and won't be formatted", col
                )
                continue

            if "int" in str(parameter.dtype).lower():
                temp[col] = pd.to_numeric(temp[col]).round(0)

            temp[col] = temp[col].astype(parameter.dtype, errors="ignore")

            if "float" in str(parameter.dtype).lower():
                temp[col] = temp[col].round(1)

        return temp

    @classmethod
    def clean(
        cls, df: pd.DataFrame, granularity: Granularity, fill=None
    ) -> pd.DataFrame:
        """
        Remove invalid data from a DataFrame
        """
        temp = copy(df)

        for col in temp.columns:
            if "_source" in col:
                continue

            parameter = parameter_service.get_parameter(col, granularity)

            if not parameter:
                logger.warning(
                    "Column %s is not a valid column name and won't be cleaned", col
                )
                continue

            for validator in parameter.validators:
                test = cls._apply_validator(validator, temp, col)
                temp.loc[~test, col] = fill

        return temp

    def humanize(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Convert wind direction and condition codes to human-readable values
        """
        temp = copy(df)

        if Parameter.WDIR in temp.columns:
            temp[Parameter.WDIR] = temp[Parameter.WDIR].apply(to_direction)

        if Parameter.COCO in temp.columns:
            temp[Parameter.COCO] = temp[Parameter.COCO].apply(to_condition)

        return temp

    @classmethod
    def convert(cls, df, granularity: Granularity, units: UnitSystem) -> pd.DataFrame:
        """
        Convert units in a DataFrame
        """
        temp = copy(df)

        for col in temp.columns:
            if "_source" in col:
                continue

            parameter = parameter_service.get_parameter(col, granularity)

            if not parameter:
                logger.warning(
                    "Column %s is not a valid column name and won't be converted", col
                )
                continue

            if parameter.unit in CONVERSION_MAPPINGS:
                if units in CONVERSION_MAPPINGS[parameter.unit]:
                    conversion_func = CONVERSION_MAPPINGS[parameter.unit][units]
                    temp[col] = temp[col].apply(conversion_func)

        return temp


schema_service = SchemaService()



"""
Data Service

The Data Service is responsible for fetching meteorological data from
different providers and merging it into a single time series.
"""

from datetime import datetime
from typing import List, Optional, Union, cast

import pandas as pd

from meteostat.api.timeseries import TimeSeries
from meteostat.core.logger import logger
from meteostat.core.parameters import parameter_service
from meteostat.core.providers import provider_service
from meteostat.core.schema import schema_service
from meteostat.enumerations import Parameter, Provider
from meteostat.typing import Station, Request
from meteostat.utils.data import stations_to_df
from meteostat.utils.guards import request_size_guard


class DataService:
    """
    Data Service
    """

    @staticmethod
    def _add_source(df: pd.DataFrame, provider_id: str) -> pd.DataFrame:
        """
        Add source column to DataFrame
        """
        if "source" not in df.index.names:
            df["source"] = provider_id
            df = df.set_index(["source"], append=True)

        return df

    @staticmethod
    def filter_time(
        df: pd.DataFrame,
        start: Union[datetime, None] = None,
        end: Union[datetime, None] = None,
    ) -> pd.DataFrame:
        """
        Filter time series data based on start and end date
        """

        # Return empty DataFrame if input is empty
        if df.empty:
            return df

        # Get time index
        time = df.index.get_level_values("time")

        # Filter & return
        try:
            return df.loc[(time >= start) & (time <= end)] if start and end else df
        except TypeError:
            return (
                df.loc[(time >= start.date()) & (time <= end.date())]
                if start and end
                else df
            )

    @staticmethod
    def concat_fragments(
        fragments: List[pd.DataFrame],
        parameters: List[Parameter],
    ) -> pd.DataFrame:
        """
        Concatenate multiple fragments into a single DataFrame
        """
        try:
            cleaned = [
                df.dropna(how="all", axis=1) if not df.empty else None
                for df in fragments
            ]
            filtered = [df for df in cleaned if df is not None]
            if not filtered:
                return pd.DataFrame()
            df = pd.concat(filtered)
            df = schema_service.fill(df, parameters)
            df = schema_service.purge(df, parameters)
            return df
        except ValueError:
            return pd.DataFrame()

    def _fetch_provider_data(
        self, req: Request, station: Station, provider: Provider
    ) -> Optional[pd.DataFrame]:
        """
        Fetch data for a single weather station and provider
        """
        try:
            # Fetch DataFrame for current provider
            df = provider_service.fetch_data(provider, req, station)

            # Continue if no data was returned
            if df is None:
                return None

            # Add current station ID to DataFrame
            df = pd.concat([df], keys=[station.id], names=["station"])

            # Add source index column to DataFrame
            df = self._add_source(df, provider)

            # Filter DataFrame for requested parameters and time range
            df = self.filter_time(df, req.start, req.end)

            # Drop empty rows
            df = df.dropna(how="all")

            return df

        except Exception:
            logger.error(
                'Could not fetch data for provider "%s"',
                provider,
                exc_info=True,
            )

    def _fetch_station_data(self, req: Request, station: Station) -> List[pd.DataFrame]:
        """
        Fetch data for a single weather station
        """
        fragments = []

        filtered_providers = provider_service.filter_providers(req, station)

        for provider in filtered_providers:
            df = self._fetch_provider_data(req, station, provider)

            # Continue if no data was returned
            if df is None:
                continue

            fragments.append(df)

        return fragments

    def fetch(
        self,
        req: Request,
    ) -> TimeSeries:
        """
        Load meteorological time series data from different providers
        """
        # Guard request
        request_size_guard(req)

        # Convert stations to list if single Station
        stations: List[Station] = (
            cast(List[Station], req.station)
            if isinstance(req.station, list)
            else [req.station]
        )

        logger.debug(
            "%s time series requested for %s station(s)", req.granularity, len(stations)
        )

        # Filter parameters
        req.parameters = parameter_service.filter_parameters(
            req.granularity, req.parameters
        )

        fragments = []

        # Go through all weather stations
        for station in stations:
            station_fragments = self._fetch_station_data(req, station)

            if station_fragments:
                fragments.extend(station_fragments)

        # Merge data in a single DataFrame
        if fragments:
            df = self.concat_fragments(fragments, req.parameters)
        else:
            df = pd.DataFrame()

        # Set data types
        df = schema_service.format(df, req.granularity)

        # Create time series
        ts = TimeSeries(
            req.granularity,
            stations_to_df(stations),
            df,
            req.start,
            req.end,
            req.timezone,
            multi_station=isinstance(req.station, list),
        )

        # Return time series
        return ts


data_service = DataService()



"""
Configuration Service

Manages configuration settings for Meteostat, including cache, network,
stations, interpolation, and provider-specific settings. Configuration can be
loaded from environment variables with the MS_ prefix.
"""

import os
import json
from typing import Any, List, Optional

from meteostat.core.logger import logger
from meteostat.enumerations import TTL, Parameter
from meteostat.utils.types import extract_property_type, validate_parsed_value


class ConfigService:
    """
    Configuration Service for Meteostat
    """

    prefix: str

    # Sentinel value to indicate we should skip this env var (distinct from None)
    _SKIP_VALUE = object()

    @property
    def _prefix(self) -> str:
        """
        The environment variable prefix
        """
        return f"{self.prefix}_" if self.prefix else ""

    def _parse_env_value(self, key: str, value: str) -> Any:
        """
        Parse an environment variable value and validate against property type.

        For string types, the value is used directly without JSON parsing.
        For other types (bool, int, list, dict), JSON parsing is used.

        If validation fails, the error is logged and a sentinel value (_SKIP_VALUE)
        is returned to skip the invalid environment variable without aborting initialization.

        Returns None if the environment variable value is explicitly set to null.
        """
        # Extract the expected type for the property
        try:
            expected_type, original_type = extract_property_type(self.__class__, key)
        except ValueError:
            # Property doesn't exist on this config class, skip it
            logger.debug(
                "Environment variable '%s' does not match any config property", key
            )
            return self._SKIP_VALUE

        if expected_type is None:
            # Fallback to JSON parsing if no type annotation is available
            try:
                return json.loads(value)
            except (json.JSONDecodeError, TypeError, ValueError):
                logger.error("Failed to parse environment variable '%s'", key)
                return self._SKIP_VALUE

        # For string types, use the value directly without JSON parsing
        if expected_type is str:
            return value

        # Parse the value using JSON for non-string types
        try:
            parsed_value = json.loads(value)
        except (json.JSONDecodeError, TypeError, ValueError):
            logger.error("Failed to parse environment variable '%s'", key)
            return self._SKIP_VALUE

        # Validate and potentially convert the parsed value
        try:
            return validate_parsed_value(
                parsed_value, expected_type, original_type, key
            )
        except (ValueError, TypeError) as e:
            logger.error(
                "Failed to validate environment variable '%s': %s", key, str(e)
            )
            return self._SKIP_VALUE

    def _set_env_value(self, key: str, value: Any) -> None:
        """
        Set a configuration using a key-value pair
        """
        setattr(self, key, value)

    def __init__(self, prefix: str = "MS") -> None:
        """
        Initialize configuration service
        """
        self.prefix = prefix
        self.load_env()

    def get_env_name(self, key: str) -> str:
        """
        Get the environment variable name for a given key
        """
        if not hasattr(self, key):
            raise KeyError(f"Configuration has no key '{key}'")

        key = f"{self._prefix}{key}"
        return key.upper()

    def load_env(self) -> None:
        """
        Update configuration from environment variables with a given prefix.
        """
        for key, value in os.environ.items():
            if not key.startswith(self._prefix):
                continue

            key = key.replace(self._prefix, "").lower()
            value = self._parse_env_value(key, value)

            # Skip invalid values (but allow None to be set explicitly)
            if value is not self._SKIP_VALUE:
                self._set_env_value(key, value)


class Config(ConfigService):
    """
    Meteostat Configuration

    Manages all configuration settings including cache, network, stations,
    interpolation, and provider-specific settings. Supports loading configuration
    from environment variables.
    """

    # General settings
    block_large_requests: bool = True  # Block requests that include too many stations

    # Cache settings
    cache_enable: bool = True
    cache_directory: str = (
        os.path.expanduser("~") + os.sep + ".meteostat" + os.sep + "cache"
    )
    cache_ttl: int = TTL.MONTH
    cache_autoclean: bool = True

    # Network settings
    network_proxies: Optional[dict] = None

    # Station meta data settings
    stations_db_ttl: int = TTL.WEEK
    stations_db_endpoints: List[str] = [
        "https://data.meteostat.net/stations.db",
        "https://raw.githubusercontent.com/meteostat/weather-stations/master/stations.db",
    ]
    stations_db_file: str = (
        os.path.expanduser("~") + os.sep + ".meteostat" + os.sep + "stations.db"
    )

    # Interpolation settings
    lapse_rate_parameters = [Parameter.TEMP, Parameter.TMIN, Parameter.TMAX]

    # [Provider] Meteostat settings
    include_model_data: bool = True
    hourly_endpoint: str = "https://data.meteostat.net/hourly/{year}/{station}.csv.gz"
    daily_endpoint: str = "https://data.meteostat.net/daily/{year}/{station}.csv.gz"
    monthly_endpoint: str = "https://data.meteostat.net/monthly/{station}.csv.gz"

    # [Provider] DWD settings
    dwd_ftp_host: str = "opendata.dwd.de"
    dwd_hourly_modes: Optional[List[str]] = None
    dwd_daily_modes: Optional[List[str]] = None
    dwd_climat_modes: Optional[List[str]] = None

    # [Provider] NOAA settings
    aviationweather_endpoint: str = (
        "https://aviationweather.gov/api/data/metar?"
        "ids={station}&format=raw&taf=false&hours=24"
    )
    aviationweather_user_agent: Optional[str] = None

    # [Provider] Met.no settings
    metno_forecast_endpoint: str = (
        "https://api.met.no/weatherapi/locationforecast/2.0/compact?"
        "lat={latitude}&lon={longitude}&altitude={elevation}"
    )
    metno_user_agent: Optional[str] = None

    # [Provider] GSA settings
    gsa_api_base_url: str = "https://dataset.api.hub.geosphere.at/v1"


config = Config("MS")



"""
Inverse Distance Weighting (IDW) Interpolation

Implements IDW interpolation for spatial weather data with support for
elevation-weighted distance calculations.
"""

from typing import Callable

import numpy as np
import pandas as pd

from meteostat.api.point import Point
from meteostat.api.timeseries import TimeSeries


def inverse_distance_weighting(
    power: float = 2.0,
) -> Callable:
    """
    Interpolate values using Inverse Distance Weighting (IDW).

    This method calculates interpolated values as a weighted average of nearby
    stations, where weights decrease with distance. Optionally incorporates
    elevation differences in the distance calculation.

    Parameters
    ----------
    power : float, optional
        Power parameter for IDW (default: 2.0). Higher values give more
        weight to closer stations.
    elevation_weight : float, optional
        Weight for elevation difference in distance calculation (default: 0.1).
        The effective distance is calculated as:
        sqrt(horizontal_distance^2 + (elevation_diff * elevation_weight)^2)

    Notes
    -----
    - If elevation data is missing for either the point or stations, only
    horizontal distance is used.
    - Stations with zero effective distance get weight of 1.0, all others get 0.
    - All numeric columns except location-related ones are interpolated.
    """

    def _get_df(
        df: pd.DataFrame,
        ts: TimeSeries,
        point: Point,
    ) -> pd.DataFrame:
        # Group by time to interpolate each timestamp separately
        grouped = df.groupby(pd.Grouper(level="time", freq=ts.freq))

        # List to store interpolated results for each time period
        interpolated_results = []

        for time_idx, group in grouped:
            if group.empty:
                continue

            effective_distance = group["effective_distance"]

            # Calculate weights using IDW formula: w = 1 / d^p
            # Handle zero distance case (station at exact location)
            min_distance = effective_distance.min()
            if min_distance == 0:
                # If any station is at the exact location, use only that station
                weights = (effective_distance == 0).astype(float)
            else:
                # Standard IDW weights
                weights = 1.0 / (effective_distance**power)

            # Normalize weights so they sum to 1
            weights = weights / weights.sum()

            # Get numeric columns to interpolate (exclude location-related columns)
            location_cols = ["latitude", "longitude", "elevation", "distance"]
            numeric_cols = [
                col
                for col in group.columns
                if col not in location_cols
                and pd.api.types.is_numeric_dtype(group[col])
            ]

            # Calculate weighted average for each numeric column
            interpolated_row = {}
            for col in numeric_cols:
                # Only use non-NaN values for interpolation
                valid_mask = group[col].notna()
                if valid_mask.any():
                    valid_values = group.loc[valid_mask, col]
                    valid_weights = weights[valid_mask]
                    # Re-normalize weights for valid values only
                    valid_weights = valid_weights / valid_weights.sum()
                    interpolated_row[col] = (valid_values * valid_weights).sum()
                else:
                    # If all values are NaN, result is NaN
                    interpolated_row[col] = np.nan

            # Add location information from the point
            interpolated_row["latitude"] = point.latitude
            interpolated_row["longitude"] = point.longitude
            if point.elevation is not None:
                interpolated_row["elevation"] = point.elevation
            interpolated_row["distance"] = 0  # Distance from point to itself

            # Create a DataFrame row with the time index
            result_df = pd.DataFrame(
                [interpolated_row], index=pd.DatetimeIndex([time_idx])
            )
            result_df.index.name = "time"
            interpolated_results.append(result_df)

        # Combine all time periods
        if interpolated_results:
            result = pd.concat(interpolated_results)
            return result
        else:
            return pd.DataFrame()

    return _get_df



"""
GeoSphere Austria Data Hub daily data import routine

Get daily climate data for weather stations in Austria.

License: CC BY 4.0 (https://creativecommons.org/licenses/by/4.0/)
"""

from datetime import datetime
from typing import Dict, Optional

import pandas as pd

from meteostat.enumerations import TTL, Parameter
from meteostat.core.logger import logger
from meteostat.typing import ProviderRequest
from meteostat.core.cache import cache_service
from meteostat.api.config import config
from meteostat.core.network import network_service
from meteostat.utils.conversions import (
    hours_to_minutes,
    ms_to_kmh,
    percentage_to_okta,
    pres_to_msl,
)


RESOURCE_ID = "klima-v2-1d"

# Mapping from GeoSphere Austria parameter names to Meteostat parameters
# See: https://dataset.api.hub.geosphere.at/v1/station/historical/klima-v2-1d/metadata
PARAMETER_MAPPING: Dict[str, Parameter] = {
    "tl_mittel": Parameter.TEMP,  # Mean air temperature (°C) - Lufttemperatur 2m Mittelwert
    "tlmax": Parameter.TMAX,  # Maximum air temperature (°C) - Lufttemperatur 2m Maximalwert
    "tlmin": Parameter.TMIN,  # Minimum air temperature (°C) - Lufttemperatur 2m Minimalwert
    "rr": Parameter.PRCP,  # Precipitation 24h sum (mm) - Niederschlag 24h Summe
    "sh": Parameter.SNWD,  # Snow depth (cm) - Schneehöhe
    "vv_mittel": Parameter.WSPD,  # Mean wind speed (km/h) - Windgeschwindigkeit Mittelwert
    "ffx": Parameter.WPGT,  # Maximum wind gust (m/s) - Windböe Maximum
    "p_mittel": Parameter.PRES,  # Mean air pressure (hPa) - Luftdruck Mittelwert
    "rf_mittel": Parameter.RHUM,  # Mean relative humidity (%) - Relative Feuchte Mittelwert
    "so_h": Parameter.TSUN,  # Sunshine duration (h) - Sonnenscheindauer
    "bewm_mittel": Parameter.CLDC,  # Mean cloud cover (%) - Bedeckung Mittelwert
}

# Inverse mapping
METEOSTAT_TO_GSA = {v: k for k, v in PARAMETER_MAPPING.items()}


@cache_service.cache(TTL.WEEK, "pickle")
def get_data(
    station_id: str,
    elevation: int | None,
    parameters: list[str],
    start: datetime,
    end: datetime,
) -> Optional[pd.DataFrame]:
    """
    Fetch data from GeoSphere Austria Data Hub API
    """
    logger.debug(
        f"Fetching daily data for station '{station_id}' from {start} to {end}"
    )

    # Format dates as ISO 8601 (date only for daily data)
    start_str = start.strftime("%Y-%m-%d")
    end_str = end.strftime("%Y-%m-%d")

    # Build URL
    url = f"{config.gsa_api_base_url}/station/historical/{RESOURCE_ID}"

    # Make request
    response = network_service.get(
        url,
        params={
            "parameters": ",".join(parameters),
            "station_ids": station_id,
            "start": start_str,
            "end": end_str,
            "output_format": "geojson",
        },
    )

    if response.status_code != 200:
        logger.warning(
            f"Failed to fetch daily data for station {station_id} (status: {response.status_code})"
        )
        return None

    try:
        data = response.json()

        if not data.get("features"):
            logger.info(f"No daily data returned for station {station_id}")
            return None

        # Get timestamps array
        timestamps = data.get("timestamps")
        if not timestamps:
            logger.warning("No timestamps in daily response")
            return None

        # Extract time series data from GeoJSON response
        # New API format has timestamps at top level and data as arrays
        feature = data["features"][0]
        props = feature.get("properties", {})
        params_data = props.get("parameters", {})

        if not params_data:
            logger.info(f"No parameter data returned for station {station_id}")
            return None

        # Build DataFrame from timestamps and parameter arrays
        df_dict = {}
        for param in parameters:
            if param in params_data:
                param_info = params_data[param]
                if "data" in param_info:
                    df_dict[param] = param_info["data"]

        if not df_dict:
            return None

        # Create DataFrame with timestamps as index
        df = pd.DataFrame(df_dict)
        dt_index = pd.DatetimeIndex(pd.to_datetime(timestamps))
        df.index = dt_index.tz_localize(None)
        df.index.name = "time"

        # Sort by time
        df = df.sort_index()

        # Rename columns to Meteostat parameter names
        rename_map = {}
        for gsadh_param, meteostat_param in PARAMETER_MAPPING.items():
            if gsadh_param in df.columns:
                rename_map[gsadh_param] = meteostat_param

        df = df.rename(columns=rename_map)

        # Convert units where necessary
        if Parameter.WSPD in df.columns:
            df[Parameter.WSPD] = df[Parameter.WSPD].apply(ms_to_kmh)

        if Parameter.WPGT in df.columns:
            df[Parameter.WPGT] = df[Parameter.WPGT].apply(ms_to_kmh)

        if Parameter.TSUN in df.columns:
            df[Parameter.TSUN] = df[Parameter.TSUN].apply(hours_to_minutes)

        if Parameter.CLDC in df.columns:
            df[Parameter.CLDC] = df[Parameter.CLDC].apply(percentage_to_okta)

        if Parameter.PRES in df.columns:
            df[Parameter.PRES] = df.apply(
                lambda row: pres_to_msl(row, elevation), axis=1
            )

        # Round values
        df = df.round(1)

        return df

    except Exception as error:
        logger.warning(f"Error parsing daily response: {error}", exc_info=True)
        return None


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    """
    Fetch daily data from GeoSphere Austria Data Hub
    """
    if "national" not in req.station.identifiers:
        return None

    station_id = req.station.identifiers["national"]

    # Map Meteostat parameters to GeoSphere Austria parameters
    gsa_params = []
    for param in req.parameters:
        if param in METEOSTAT_TO_GSA:
            gsa_params.append(METEOSTAT_TO_GSA[param])

    if not gsa_params:
        logger.info("No mappable parameters for GeoSphere Austria daily data")
        return None

    # Fetch data
    df = get_data(station_id, req.station.elevation, gsa_params, req.start, req.end)

    if df is None or df.empty:
        return None

    return df



"""
Process GHCND data

This code is based on an implementation by
Aaron Penne (https://github.com/aaronpenne).

The code is licensed under the MIT license.
"""

from io import StringIO
from ftplib import FTP
from typing import Optional

from numpy import nan
import pandas as pd

from meteostat.enumerations import TTL, Parameter
from meteostat.typing import ProviderRequest
from meteostat.core.cache import cache_service
from meteostat.utils.conversions import ms_to_kmh, percentage_to_okta

FTP_SERVER = "ftp.ncdc.noaa.gov"
COLUMN_NAMES = {
    "MM/DD/YYYY": "time",
    "TMAX": Parameter.TMAX,
    "TMIN": Parameter.TMIN,
    "TAVG": Parameter.TEMP,
    "PRCP": Parameter.PRCP,
    "SNWD": Parameter.SNWD,
    "AWDR": Parameter.WDIR,
    "AWND": Parameter.WSPD,
    "TSUN": Parameter.TSUN,
    "WSFG": Parameter.WPGT,
    "ACMC": Parameter.CLDC,
}


def connect_to_ftp():
    """
    Connect to FTP server
    """
    ftp = FTP(FTP_SERVER)
    ftp.login()

    return ftp


def get_flags(string):
    """
    Get flags, replacing empty flags with '_' for clarity (' S ' becomes '_S_')
    """
    m_flag = string.read(1)
    m_flag = m_flag if m_flag.strip() else "_"
    q_flag = string.read(1)
    q_flag = q_flag if q_flag.strip() else "_"
    s_flag = string.read(1)
    s_flag = s_flag if s_flag.strip() else "_"

    return [m_flag + q_flag + s_flag]


def create_df(element, dict_element):
    """
    Create dataframes from dicts, turn indices into date strings (YYYY-MM-DD)
    """
    element = element.upper()
    df_element = pd.DataFrame(dict_element)

    # Add dates (YYYY-MM-DD) as index on df. Pad days with zeros to two places
    df_element.index = (
        df_element["YEAR"]
        + "-"
        + df_element["MONTH"]
        + "-"
        + df_element["DAY"].str.zfill(2)
    )
    df_element.index.name = "DATE"

    # Arrange columns so ID, YEAR, MONTH, DAY are at front. Leaving them in
    # for plotting later - https://stackoverflow.com/a/31396042
    for col in ["DAY", "MONTH", "YEAR", "ID"]:
        df_element = move_col_to_front(col, df_element)

    # Convert numerical values to float
    df_element[element] = df_element[element].astype(float)

    return df_element


def move_col_to_front(element, df):
    """
    Move DataFrame column to position 0
    """
    element = element.upper()
    cols = df.columns.tolist()
    cols.insert(0, cols.pop(cols.index(element)))
    df = df.reindex(columns=cols)

    return df


# pylint: disable=too-many-locals
def dly_to_df(ftp, station_id):
    """
    Convert .dly files to DataFrame
    """
    ftp_filename = station_id + ".dly"

    # Write .dly file to stream using StringIO using FTP command 'RETR'
    stream = StringIO()
    ftp.retrlines("RETR " + "/pub/data/ghcn/daily/all/" + ftp_filename, stream.write)

    # Move to first char in file
    stream.seek(0)

    # File params
    num_chars_line = 269

    # Read through entire StringIO stream (the .dly file)
    # and collect the data
    all_dicts = {}
    element_flag = {}
    index = 0

    while True:
        index += 1

        # Read metadata for each line
        # (one month of data for a particular element per line)
        stream.read(11)  # station ID (not used in this loop)
        year = stream.read(4)
        month = stream.read(2)
        day = 0
        element = stream.read(4)

        # If this is blank then we've reached EOF and should exit loop
        if not element:
            break

        # Loop through each day in rest of row,
        # break if current position is end of row
        while stream.tell() % num_chars_line != 0:
            day += 1
            # Fill in contents of each dict depending on element type in
            # current row
            if day == 1:
                try:
                    element_flag[element]
                except BaseException:
                    element_flag[element] = 1
                    all_dicts[element] = {}
                    all_dicts[element]["ID"] = []
                    all_dicts[element]["YEAR"] = []
                    all_dicts[element]["MONTH"] = []
                    all_dicts[element]["DAY"] = []
                    all_dicts[element][element.upper()] = []
                    all_dicts[element][element.upper() + "_FLAGS"] = []

            value = stream.read(5)
            flags = get_flags(stream)
            if value == "-9999":
                continue
            all_dicts[element]["ID"] += [station_id]
            all_dicts[element]["YEAR"] += [year]
            all_dicts[element]["MONTH"] += [month]
            all_dicts[element]["DAY"] += [str(day)]
            all_dicts[element][element.upper()] += [value]
            all_dicts[element][element.upper() + "_FLAGS"] += flags

    # Create dataframes from dict
    all_dfs = {}
    for element in list(all_dicts.keys()):
        all_dfs[element] = create_df(element, all_dicts[element])

    # Combine all element dataframes into one dataframe,
    # indexed on date.
    #
    # pd.concat automagically aligns values to matching indices, therefore the
    # data is date aligned!
    list_dfs = []
    for df in list(all_dfs.keys()):
        list_dfs += [all_dfs[df]]
    df_all = pd.concat(list_dfs, axis=1, sort=False)
    df_all.index.name = "MM/DD/YYYY"

    # Remove duplicated/broken columns and rows
    # https://stackoverflow.com/a/40435354
    df_all = df_all.loc[:, ~df_all.columns.duplicated()]
    df_all = df_all.loc[df_all["ID"].notnull(), :]

    return df_all


@cache_service.cache(TTL.DAY, "pickle")
def get_df(station: str) -> pd.DataFrame:
    ftp = connect_to_ftp()
    df = dly_to_df(ftp, station)
    # Filter relevant columns
    df = df.drop(columns=[col for col in df if col not in COLUMN_NAMES.keys()])
    # Add missing columns
    for col in list(COLUMN_NAMES.keys())[1:]:
        if col not in df.columns:
            df[col] = nan

    # Rename columns
    df = df.reset_index().rename(columns=COLUMN_NAMES)

    # Adapt columns
    df["time"] = pd.to_datetime(df["time"])
    df[Parameter.TEMP] = df[Parameter.TEMP].div(10)
    df[Parameter.TMIN] = df[Parameter.TMIN].div(10)
    df[Parameter.TMAX] = df[Parameter.TMAX].div(10)
    df[Parameter.PRCP] = df[Parameter.PRCP].div(10)
    df[Parameter.SNWD] = df[Parameter.SNWD].div(10)
    df[Parameter.WSPD] = df[Parameter.WSPD].div(10).apply(ms_to_kmh)
    df[Parameter.WPGT] = df[Parameter.WPGT].div(10).apply(ms_to_kmh)
    df[Parameter.CLDC] = df[Parameter.CLDC].apply(percentage_to_okta)

    return df.set_index("time")


def fetch(req: ProviderRequest) -> Optional[pd.DataFrame]:
    ghcn_id = (
        req.station.identifiers["ghcn"] if "ghcn" in req.station.identifiers else None
    )
    if not ghcn_id:
        return None
    return get_df(ghcn_id)



import functools
from urllib.error import HTTPError
from typing import Optional, Callable, TypeVar

import pandas as pd

from meteostat.api.config import config
from meteostat.core.logger import logger
from meteostat.core.providers import provider_service
from meteostat.enumerations import Grade

T = TypeVar("T")


def _get_station_year_info(args: tuple) -> str:
    """
    Get station and year info
    """
    station = args[0] if len(args) > 0 else None
    year = args[1] if len(args) > 1 else None

    info = "no station or year info"

    if station is not None:
        info = f'station "{station}"'
    if year is not None:
        info += f' and year "{year}"'

    return info


def handle_exceptions(func: Callable[..., Optional[T]]) -> Callable[..., Optional[T]]:
    """
    Decorator to handle exceptions during data fetching
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Optional[T]:
        try:
            return func(*args, **kwargs)
        except HTTPError as error:
            if error.code == 404:
                logger.info(
                    f"Data file for {_get_station_year_info(args)} was not found"
                )
            else:
                logger.warning(
                    f"HTTP error while loading data file for {_get_station_year_info(args)} (status: {error.code})",
                    exc_info=True,
                )
        except Exception:
            logger.warning(
                f"Could not load data file for {_get_station_year_info(args)}",
                exc_info=True,
            )
        return None

    return wrapper


def filter_model_data(func: Callable[..., Optional[T]]) -> Callable[..., Optional[T]]:
    """
    Decorator to filter out model/forecast data based on configuration
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Optional[T]:
        result = func(*args, **kwargs)

        if (
            not config.include_model_data
            and isinstance(result, pd.DataFrame)
            and result is not None
        ):
            logger.debug("Filtering out model/forecast data")

            excluded_providers = [
                provider.id
                for provider in provider_service.providers
                if provider.grade
                in (
                    Grade.FORECAST,
                    Grade.ANALYSIS,
                )
            ]

            mask = result.index.get_level_values("source").isin(excluded_providers)

            result = result[~mask]

        return result

    return wrapper



"""
Functions for converting between different
meteorological data units.

The code is licensed under the MIT license.
"""

import math
from typing import Optional

import pandas as pd

from meteostat.enumerations import Parameter, Unit, UnitSystem


def celsius_to_fahrenheit(value):
    """
    Convert Celsius to Fahrenheit
    """

    return round((value * 9 / 5) + 32, 1)


def celsius_to_kelvin(value):
    """
    Convert Celsius to Kelvin
    """

    return round(value + 273.15, 1)


def millimeters_to_inches(value):
    """
    Convert millimeters to inches
    """

    return round(value / 25.4, 3)


def centimeters_to_inches(value):
    """
    Convert centimeters to inches
    """

    return round(value / 2.54, 3)


def meters_to_feet(value):
    """
    Convert meters to feet
    """

    return round(value / 0.3048, 1)


def kmh_to_ms(value):
    """
    Convert kilometers per hour to meters per second
    """

    return round(value / 3.6, 1)


def kmh_to_mph(value):
    """
    Convert kilometers per hour to miles per hour
    """

    return round(value * 0.6214, 1)


def kelvin_to_celsius(value):
    """
    Convert Kelvin to Celsius
    """
    return value - 273.15 if not pd.isna(value) else None


def ms_to_kmh(value):
    """
    Convert m/s to km/h
    """
    return value * 3.6 if not pd.isna(value) else None


def hours_to_minutes(value):
    """
    Convert duration from hours to minutes
    """
    return value * 60 if not pd.isna(value) else None


def temp_dwpt_to_rhum(row: dict):
    """
    Get relative humidity from temperature and dew point
    """
    return (
        100
        * (
            math.exp((17.625 * row["dwpt"]) / (243.04 + row["dwpt"]))
            / math.exp((17.625 * row["temp"]) / (243.04 + row["temp"]))
        )
        if not pd.isna(row["temp"]) and not pd.isna(row["dwpt"])
        else None
    )


def pres_to_msl(row: dict, altitude: Optional[int] = None, temp: str = Parameter.TEMP):
    try:
        pres = row.get(Parameter.PRES)
        t = row.get(temp)

        if pd.isna(pres) or pd.isna(t) or altitude is None or pres == -999:
            return None

        # Type narrowing for arithmetic operations
        if not isinstance(pres, (int, float)) or not isinstance(t, (int, float)):
            return None

        return round(
            pres
            * math.pow(
                1 - (0.0065 * altitude) / (t + 0.0065 * altitude + 273.15),
                -5.257,
            ),
            1,
        )
    except Exception:
        return None


def percentage_to_okta(value):
    """
    Convert cloud cover percentage to oktas
    """
    return round(value / 12.5) if not pd.isna(value) else None


def jcm2_to_wm2(value):
    """
    Convert Joule/CM^2 to Watt/M^2
    """
    return round(value * 2.78) if not pd.isna(value) else None


def to_direction(value):
    """
    Convert degrees to wind direction
    """

    wdir = None

    if (337 <= value <= 360) or value <= 23:
        wdir = "N"
    if 24 <= value <= 68:
        wdir = "NE"
    if 69 <= value <= 113:
        wdir = "E"
    if 114 <= value <= 158:
        wdir = "SE"
    if 159 <= value <= 203:
        wdir = "S"
    if 204 <= value <= 248:
        wdir = "SW"
    if 249 <= value <= 293:
        wdir = "W"
    if 294 <= value <= 336:
        wdir = "NW"

    return wdir


def to_condition(value):
    """
    Convert Meteostat condition code to descriptive string
    """

    if not value or value < 1 or value > 27:
        return None

    return [
        "Clear",
        "Fair",
        "Cloudy",
        "Overcast",
        "Fog",
        "Freezing Fog",
        "Light Rain",
        "Rain",
        "Heavy Rain",
        "Freezing Rain",
        "Heavy Freezing Rain",
        "Sleet",
        "Heavy Sleet",
        "Light Snowfall",
        "Snowfall",
        "Heavy Snowfall",
        "Rain Shower",
        "Heavy Rain Shower",
        "Sleet Shower",
        "Heavy Sleet Shower",
        "Snow Shower",
        "Heavy Snow Shower",
        "Lightning",
        "Hail",
        "Thunderstorm",
        "Heavy Thunderstorm",
        "Storm",
    ][int(value) - 1]


CONVERSION_MAPPINGS = {
    Unit.CELSIUS: {
        UnitSystem.IMPERIAL: celsius_to_fahrenheit,
        UnitSystem.SI: celsius_to_kelvin,
    },
    Unit.MILLIMETERS: {
        UnitSystem.IMPERIAL: millimeters_to_inches,
    },
    Unit.CENTIMETERS: {
        UnitSystem.IMPERIAL: centimeters_to_inches,
    },
    Unit.KMH: {
        UnitSystem.SI: kmh_to_ms,
        UnitSystem.IMPERIAL: kmh_to_mph,
    },
    Unit.METERS: {
        UnitSystem.IMPERIAL: meters_to_feet,
    },
}
